from functools import cached_property
from typing import Optional
import numpy as np
from dataclasses import dataclass
import plotly.graph_objects as go

def get_line(
    y: np.ndarray,
    point1: np.ndarray,
    point2: np.ndarray,
    y_int: Optional[int] = None
):
    x2, y2 = point2[0], point2[1]
    x1, y1 = point1[0], point1[1]
    m = (y2 - y1) / (x2 - x1)
    b = y2 - m*x2 if y_int is None else 0
    x = (y-b)/m
    return np.array([x, y]).T


def get_arc(
    lower_point: np.ndarray,
    upper_point: np.ndarray,
    radius: float,
    center: np.ndarray,
    num_points: int,
    is_clockwise: bool = True,
    endpoint: bool = True
):
    lower_distance = lower_point - center
    upper_distance = upper_point - center
    angle1 = np.arctan2(lower_distance[1], lower_distance[0])
    angle2 = np.arctan2(upper_distance[1], upper_distance[0])

    if not is_clockwise:
        angle1 = angle1 + 2*np.pi

    angle = np.linspace(angle1, angle2, num_points, endpoint=endpoint)
    return center + np.array([
        radius * np.cos(angle),
        radius * np.sin(angle),
    ]).T


@dataclass
class FirtreeAttachment:
    gamma: float
    "angle of upper flank line (rad)"

    beta: float
    "angle of lower flank line (rad)"

    ll: float
    "lower flank line length (m)"

    lu: float
    "upper flank line length (m)"

    Ri: float
    "inner circle radius"

    Ro: float
    "outer circle radius"

    R_dove: float
    "dove circle radius"

    max_length: float
    "max length of attachment"

    num_stages: int
    "number of stages"

    disk_radius: float
    "disk radius of attachment"

    tolerance: float
    "attachment side tolerance"

    include_top_arc: bool = True
    "whether or not to include top arc"

    num_arc_points: int = 20
    "number of arc points"

    def __post_init__(self):
        self.origin = np.array([0, 0])

        # Outer Circle
        self.outer_circle_lower_tangent = self.origin + (
            np.array([self.ll*np.cos(self.beta), self.ll*np.sin(self.beta)])
        )
        self.outer_circle_upper_tangent = self.outer_circle_lower_tangent + (
            np.array([0, 2*self.Ro*np.cos(self.gamma)])
        )
        self.outer_circle_tanget_intersect = self.outer_circle_lower_tangent + (
            np.array([self.Ro*(1 - np.sin(self.gamma)**2)/np.sin(self.gamma), self.Ro*np.cos(self.gamma),])
        )
        self.outer_circle_center = self.outer_circle_lower_tangent + (
            np.array([-self.Ro*np.sin(self.gamma), self.Ro*np.cos(self.gamma)])
        )

        # Inner Circle
        self.inner_circle_lower_tangent = self.outer_circle_upper_tangent + (
            np.array([-self.lu*np.cos(self.gamma), self.lu*np.sin(self.gamma)])
        )
        self.inner_circle_upper_tangent = self.inner_circle_lower_tangent + (
            np.array([0, 2*self.Ri*np.cos(self.gamma)])
        )
        self.inner_circle_center = self.inner_circle_lower_tangent + (
            np.array([self.Ri*np.sin(self.beta), self.Ri*np.cos(self.beta)])
        )

        # Dove
        self.dove_circle_center = np.array([self.R_dove*np.sin(self.beta), -self.R_dove*np.cos(self.beta)])
        self.dove_lower_point = self.dove_circle_center + np.array([0,-self.R_dove])

    @cached_property
    def dove_arc(self):
        "calculates firtree dove arc"
        return get_arc(self.dove_lower_point, self.origin, self.R_dove, self.dove_circle_center, self.num_arc_points, is_clockwise=False)

    def get_top_shape(self, include_tolerance: bool):
        "calculates firtree top arc or line segement"
        max_length = self.max_length + self.tolerance if include_tolerance else self.max_length
        top_arc_left_point = self.left_side[-1]
        top_arc_right_point = top_arc_left_point + np.array([max_length, 0])
        
        if self.include_top_arc:
            sector_angle = 2*np.arcsin((max_length/2)/self.disk_radius)
            top_arc_height = self.disk_radius - (max_length/2)/np.tan(sector_angle/2)
            disk_center = np.array([0,top_arc_left_point[1]-self.disk_radius+top_arc_height])
            return get_arc(top_arc_left_point, top_arc_right_point, self.disk_radius, disk_center, self.num_arc_points)

        return np.concatenate([[top_arc_left_point], [top_arc_right_point]])

    def get_stage(self, end_stage: bool = False):
        "calculates firtree single stage coordinates"
        # Flank Lines
        yl = np.linspace(self.origin[1], self.outer_circle_lower_tangent[1], 2, endpoint=False)
        lower_flank_line = get_line(yl, self.origin, self.outer_circle_lower_tangent)

        yu = np.linspace(self.outer_circle_upper_tangent[1], self.inner_circle_lower_tangent[1], 2, endpoint=False)
        upper_flank_line = get_line(yu, self.outer_circle_tanget_intersect, self.outer_circle_upper_tangent)

        # Arcs
        outer_arc = get_arc(self.outer_circle_lower_tangent, self.outer_circle_upper_tangent, self.Ro, self.outer_circle_center, self.num_arc_points, endpoint=False)
        inner_arc = get_arc(self.inner_circle_lower_tangent, self.inner_circle_upper_tangent, self.Ri, self.inner_circle_center, self.num_arc_points, is_clockwise=False)
        
        stage_elements = [lower_flank_line,outer_arc, upper_flank_line]
        if not end_stage:
            stage_elements.append(inner_arc)
        
        return np.concatenate(stage_elements)
    
    @cached_property
    def left_side(self) -> np.ndarray:
        "calculates firtree attachment left side coordinates"
        stage = self.get_stage()
        attachment_stage_side:Optional[np.ndarray] = None
        assert self.num_stages > 0, "num stages must greater than 0"
        # TODO: make this more efficient with Numba
        for i in range(self.num_stages):
            next_stage = stage
            if attachment_stage_side is not None:
                attachment_stage_side = np.concatenate([
                    attachment_stage_side[:-1], 
                    next_stage + attachment_stage_side[-1]
                ])
            else:
                attachment_stage_side = next_stage
        
        # offset side to max length
        assert attachment_stage_side is not None
        attachment_center_offset = np.array([-attachment_stage_side[-1][0]-self.max_length/2, 0])
        return np.concatenate([self.dove_arc[:-1], attachment_stage_side]) + attachment_center_offset
    

    def get_coords(self, include_tolerance: bool):
        "calculates firtree attachment coordinates"
        top_arc = self.get_top_shape(include_tolerance)
        left_side = self.left_side + np.array([-self.tolerance/2, 0]) if include_tolerance else self.left_side
        attachment_right_side = np.flip(left_side, axis=0) * np.array([-1, 1])        
        attachment = np.concatenate([left_side[:-1], top_arc[:-1], attachment_right_side, [left_side[0]]])
        return attachment + np.array([0, -np.max(attachment[:,1])])

    @cached_property
    def coords(self):
        return self.get_coords(include_tolerance=False)

    @cached_property
    def coords_with_tolerance(self):
        return self.get_coords(include_tolerance=True)

    @cached_property
    def height(self):
        return np.max(self.coords[:, 1]) - np.min(self.coords[:, 1]) # type: ignore

    @cached_property
    def bottom_width(self):
        return np.abs(self.left_side[0][0])*2

    def visualize(self):
        fig = go.Figure()
        fig.add_trace(go.Scatter(
            x=self.coords[:, 0],
            y=self.coords[:, 1],
        ))
        fig.layout.yaxis.scaleanchor = "x"
        fig.show()




"""
TurboDesigner - The open-source turbomachinery designer
"""

__version__ = "0.1.0"

# Import main modules for easier access
from turbodesigner.flow_station import FlowStation
from turbodesigner.stage import Stage
from turbodesigner.turbomachinery import Turbomachinery
from turbodesigner.units import MM, DEG, BAR
from turbodesigner.visualizer import TurbomachineryVisualizer
from turbodesigner.exporter import TurbomachineryExporter

# Make these modules available at the package level
__all__ = [
    "FlowStation",
    "Stage",
    "Turbomachinery",
    "MM",
    "DEG",
    "BAR",
    "TurbomachineryVisualizer",
    "TurbomachineryExporter",
]


from dataclasses import dataclass
from functools import cached_property
from typing import Union
import numpy as np
from turbodesigner.airfoils import AirfoilType
from turbodesigner.blade.metal_angles import MetalAngles


@dataclass
class MetalAngleOffset:
    i: Union[float, np.ndarray]
    "blade incidence (rad)"

    delta: Union[float, np.ndarray]
    "blade deviation (rad)"


@dataclass
class JohnsenBullockMetalAngleMethod:
    "Johnsen and Bullock 1965"

    beta1: Union[float, np.ndarray]
    "inlet flow angle (rad)"

    beta2: Union[float, np.ndarray]
    "outlet flow angle (rad)"

    sigma: float
    "spacing between blades (m)"

    tbc: float
    "max thickness to chord (dimensionless)"

    airfoil_type: AirfoilType
    "airfoil type (AirfoilType)"

    def __post_init__(self):
        self.beta1_deg = np.abs(np.degrees(self.beta1))
        self.beta2_deg = np.abs(np.degrees(self.beta2))

    @cached_property
    def Ksh(self):
        "blade shape paramter (dimensionless)"
        if self.airfoil_type == AirfoilType.NACA65:
            return 1.0
        elif self.airfoil_type == AirfoilType.DCA:
            return 0.7
        elif self.airfoil_type == AirfoilType.C4:
            return 1.1
        else:
            return 1.0  # Default case

    @cached_property
    def n(self):
        "slope factor (dimensionless)"
        return 0.025*self.sigma - ((1/90)*self.beta1_deg)**(1.2*self.sigma + 1)/(0.43*self.sigma + 1.5) - 0.06

    @cached_property
    def kti(self):
        "design incidence angle correction factor (dimensionless)"
        q = 0.28/(self.tbc**0.3 + 0.1)
        return (10*self.tbc)**q

    @cached_property
    def m(self):
        "deviation slope factor (dimensionless)"
        x = (1/100)*self.beta1_deg

        if self.airfoil_type == AirfoilType.NACA65:
            m1 = 0.333*x**2 - 0.0333*x + 0.17
        else:
            m1 = 0.316*x**3 - 0.132*x**2 + 0.074*x + 0.249

        b = -0.85*x**3 - 0.17*x + 0.9625
        return self.sigma**(-b)*m1

    @cached_property
    def Ktdelta(self):
        "design deviation angle correction factor (dimensionless)"
        return 37.5*self.tbc**2 + 6.25*self.tbc

    @cached_property
    def delta_star_0_10(self):
        "nominal deviation angle theta=0, tbc=0.10 (deg)"
        return 0.01*self.beta1_deg*self.sigma + ((0.74*self.sigma**1.9) + 3*self.sigma)*(self.beta1_deg/90)**(1.09*self.sigma + 1.67)

    @cached_property
    def i_star_0_10(self):
        "nominal incidence angle theta=0, tbc=0.10 (deg)"
        # seal pitch (dimensionless)
        p = (1/160)*self.sigma**3 + 0.914
        return ((self.beta1_deg**p)/(5 + 46*np.exp(-2.3*self.sigma))) - 0.1*self.sigma**3*np.exp((self.beta1_deg - 70)/4)

    def get_i_star_deg(self, theta_deg: Union[float, np.ndarray]):
        "nominal incidence angle (deg)"
        return theta_deg*self.n + self.kti*self.i_star_0_10*self.Ksh

    def get_delta_star_deg(self, theta_deg: Union[float, np.ndarray]):
        "nominal deviation angle (deg)"
        return self.Ksh*self.Ktdelta*self.delta_star_0_10 + theta_deg*self.m

    def get_metal_angle_offset(self, iterations: int):
        i_star_deg, delta_star_deg = 0, 0
        # TODO: make this more efficient with Numba
        for _ in range(iterations):
            metal_angles_deg = MetalAngles(self.beta1_deg, self.beta2_deg, i_star_deg, delta_star_deg)
            theta_deg = metal_angles_deg.theta
            i_star_deg = self.get_i_star_deg(theta_deg)
            delta_star_deg = self.get_delta_star_deg(theta_deg)

        i = np.radians(i_star_deg) * np.sign(self.beta1)
        delta = np.radians(delta_star_deg) * np.sign(self.beta2)
        return MetalAngleOffset(i, delta)



from functools import cached_property
from dataclasses import dataclass
import json
from typing import Literal, Union
import numpy as np
from turbodesigner.blade.row import MetalAngleMethods
from turbodesigner.flow_station import FlowStation
from turbodesigner.stage import Stage, StageBladeProperty, StageCadExport
from dacite.core import from_dict
Number = Union[int, float]


@dataclass
class TurbomachineryCadExport:
    stages: list[StageCadExport]
    "turbomachinery stages"


@dataclass
class Turbomachinery:
    gamma: float
    "ratio of specific heats (dimensionless)"

    cx: float
    "inlet flow velocity in the radial and axial plane (m/s)"

    N: float
    "rotational speed (rpm)"

    Rs: float
    "specific gas constant (J/(kg*K))"

    mdot: float
    "mass flow rate (kg/s)"

    PR: float
    "pressure ratio (dimensionless)"

    Pt: float
    "ambient total pressure (Pa)"

    Tt: float
    "ambient total temperature (K)"

    eta_isen: float
    "isentropic efficiency (dimensionless)"

    N_stg: int
    "number of stages (dimensionless)"

    B_in: Union[Number, list[Number]]
    "inlet blockage factor (dimensionless)"

    B_out: Union[Number, list[Number]]
    "outlet blockage factor (dimensionless)"

    ht: float
    "hub to tip ratio (dimensionless)"

    N_stream: int
    "number of streams per blade (dimensionless)"

    Delta_T0_stg: Union[list[float], Literal["equal"]]
    "array of stage stagnation temperature rises between inlet and outlet (K)"

    R_stg: Union[Number, list[Number]]
    "array of stage reaction rates (dimensionless)"

    rgc: Union[Number, list[Number]]
    "row gap to chord (dimensionless)"

    sgc: Union[Number, list[Number]]
    "stage gap to chord (dimensionless)"

    AR: Union[StageBladeProperty, list[StageBladeProperty]]
    "aspect ratios (dimensionless)"

    sc: Union[StageBladeProperty, list[StageBladeProperty]]
    "spacing to chord ratios (dimensionless)"

    tbc: Union[StageBladeProperty, list[StageBladeProperty]]
    "max thickness to chords (dimensionless)"

    metal_angle_method: MetalAngleMethods = "JohnsenBullock"
    "metal angle method"

    @cached_property
    def Tt2(self):
        "outlet stagnation temperature (K)"
        return self.Tt*self.PR**((self.gamma - 1)/(self.eta_poly*self.gamma))

    @cached_property
    def Pt2(self):
        "stagnation outlet pressure (Pa)"
        return self.Pt*self.PR

    @cached_property
    def eta_poly(self):
        "polytropic efficiency (dimensionless)"
        return (self.gamma - 1)*np.log(self.PR)/(self.gamma*np.log((self.eta_isen + self.PR**((self.gamma - 1)/self.gamma) - 1)/self.eta_isen))

    @cached_property
    def inlet_flow_station(self):
        "inlet flow station (FlowStation)"
        flow_station = FlowStation(
            gamma=self.gamma,
            Rs=self.Rs,
            Tt=self.Tt,
            Pt=self.Pt,
            Vm=self.cx,
            mdot=self.mdot,
            B=self.B_in[0] if isinstance(self.B_in, list) else self.B_in,
            N=self.N,
        )
        flow_station.set_radius(self.ht)
        return flow_station

    @cached_property
    def outlet_flow_station(self):
        "outlet flow station (FlowStation)"
        return FlowStation(
            gamma=self.gamma,
            Rs=self.Rs,
            Tt=self.Tt2,
            Pt=self.Pt2,
            Vm=self.cx,
            mdot=self.mdot,
            B=self.B_out[-1] if isinstance(self.B_out, list) else self.B_out,
            N=self.N,
            radius=self.inlet_flow_station.radius
        )

    @cached_property
    def Delta_T0(self):
        "stagnation temperature change between outlet and inlet (dimensionless)"
        return self.outlet_flow_station.Tt - self.inlet_flow_station.Tt

    @cached_property
    def TR(self):
        "stagnation temperature ratio between outlet and inlet (dimensionless)"
        return self.outlet_flow_station.Tt/self.inlet_flow_station.Tt

    @cached_property
    def stages(self):
        "turbomachinery stages (list[Stage])"
        if isinstance(self.Delta_T0_stg, list):
            assert len(self.Delta_T0_stg) == self.N_stg, "Delta_T0 quantity does not equal N_stg"
        else:
            assert self.Delta_T0_stg == "equal", f"'{self.Delta_T0_stg}' for  Delta_T0_stg is invalid"

        if isinstance(self.R_stg, list):
            assert len(self.R_stg) == self.N_stg, "R quantity does not equal N_stg"
            assert self.R_stg[self.N_stg-1] == 0.5, "Last stage reaction only supports R=0.5"
        else:
            assert self.R_stg == 0.5, "Last stage reaction only supports R=0.5"

        if isinstance(self.AR, list):
            assert len(self.AR) == self.N_stg, "AR quantity does not equal N_stg"
        if isinstance(self.sc, list):
            assert len(self.sc) == self.N_stg, "sc quantity does not equal N_stg"
        if isinstance(self.tbc, list):
            assert len(self.tbc) == self.N_stg, "tbc quantity does not equal N_stg"

        previous_flow_station = self.inlet_flow_station
        stages: list[Stage] = []
        # TODO: make this more efficient with Numba
        for i in range(self.N_stg):
            stage = Stage(
                stage_number=i+1,
                Delta_Tt=self.Delta_T0_stg[i] if isinstance(self.Delta_T0_stg, list) else self.Delta_T0/self.N_stg,
                R=self.R_stg[i] if isinstance(self.R_stg, list) else self.R_stg,
                previous_flow_station=previous_flow_station,
                eta_poly=self.eta_poly,
                N_stream=self.N_stream,
                AR=self.AR[i] if isinstance(self.AR, list) else self.AR,
                sc=self.sc[i] if isinstance(self.sc, list) else self.sc,
                tbc=self.tbc[i] if isinstance(self.tbc, list) else self.tbc,
                rgc=self.rgc[i] if isinstance(self.rgc, list) else self.rgc,
                sgc=self.sgc[i] if isinstance(self.sgc, list) else self.sgc,
                metal_angle_method=self.metal_angle_method
            )
            previous_flow_station = stage.mid_flow_station
            if i > 0 and i < self.N_stg:
                stages[i-1].next_stage = stage
            stages.append(stage)

        return stages

    def to_cad_export(self):
        return TurbomachineryCadExport(
            stages=[
                stage.to_cad_export() for stage in self.stages
            ]
        )

    @staticmethod
    def from_dict(obj) -> "Turbomachinery":
        return from_dict(data_class=Turbomachinery, data=obj)

    @staticmethod
    def from_file(file_name: str) -> "Turbomachinery":
        with open(file_name, "r") as fp:
            obj = json.load(fp)
        return from_dict(data_class=Turbomachinery, data=obj)



from enum import Enum
import numpy as np

class AirfoilType(Enum):
    NACA65 = 1
    DCA = 2
    C4 = 2

def get_staggered_coords(coords: np.ndarray, stagger_angle: float):
    x = coords[:, 0]
    y = coords[:, 1]

    return np.array([
        x*np.cos(stagger_angle) - y*np.sin(stagger_angle),
        x*np.sin(stagger_angle) + y*np.cos(stagger_angle),
    ]).T 



"""
TurboDesigner Blade Module - Contains blade design functionality
"""

from turbodesigner.blade.metal_angles import *
from turbodesigner.blade.row import *

__all__ = [
    "metal_angles",
    "row",
]


from enum import Enum
from functools import cached_property
from dataclasses import dataclass
from typing import Literal, Optional
from turbodesigner.airfoils import AirfoilType, DCAAirfoil
from turbodesigner.blade.metal_angle_methods.johnsen_bullock import JohnsenBullockMetalAngleMethod
from turbodesigner.blade.metal_angles import MetalAngles
from turbodesigner.blade.vortex.common import Vortex
from turbodesigner.flow_station import FlowStation
from turbodesigner.attachments.firetree import FirtreeAttachment
import numpy as np
import numpy.typing as npt
from turbodesigner.units import MM
MetalAngleMethods = Literal["EqualsFlowAngles", "JohnsenBullock"]


@dataclass
class BladeRowCadExport:
    stage_number: int
    "stage number"

    disk_height: float
    "disk height (mm)"

    hub_radius: float
    "blade hub radius (mm)"

    tip_radius: float
    "blade hub radius (mm)"

    radii: npt.NDArray[np.float64]
    "blade station radius (mm)"

    airfoils: np.ndarray
    "airfoil coordinates for each blade radius (mm)"

    attachment: np.ndarray
    "attachment coordinates (mm)"

    attachment_with_tolerance: np.ndarray
    "attachment coordinates (mm)"

    attachment_height: float
    "attachment height (mm)"

    attachment_bottom_width: float
    "attachment bottom width (mm)"

    number_of_blades: int
    "number of blades"

    twist_angle: int
    "twist angle of blade"

    is_rotating: bool
    "whether blade is rotating or not"


@dataclass
class BladeRow:
    "calculates turbomachinery blade row"

    stage_number: int
    "stage number of blade row"

    stage_flow_station: FlowStation
    "blade stage flow station (FlowStation)"

    vortex: Vortex
    "blade vortex calculation for stagger angles"

    AR: float
    "aspect ratio (dimensionless)"

    sc: float
    "spacing to chord ratio (dimensionless)"

    tbc: float
    "max thickness to chord (dimensionless)"

    is_rotating: bool
    "whether blade is rotating or not"

    N_stream: int
    "number of streams per blade (dimensionless)"

    metal_angle_method: MetalAngleMethods
    "metal angle method"

    next_stage_flow_station: Optional["FlowStation"] = None
    "next blade row flow station"

    deviation_iterations: int = 20
    "nominal deviation iterations"

    def __post_init__(self):
        assert self.N_stream % 2 != 0, "N_stream must be an odd number"
        if self.is_rotating and self.next_stage_flow_station is None:
            self.next_stage_flow_station = self.stage_flow_station.copyStream(
                alpha=self.vortex.alpha(self.radii, is_rotating=False),
                radius=self.radii
            )

    @cached_property
    def rt(self):
        "blade tip radius (m)"
        rt = self.stage_flow_station.outer_radius
        assert isinstance(rt, float)
        return rt

    @cached_property
    def rh(self):
        "blade hub radius (m)"
        rh = self.stage_flow_station.inner_radius
        assert isinstance(rh, float)
        return rh

    @cached_property
    def rm(self):
        "blade mean radius (m)"
        rm = self.stage_flow_station.radius
        assert isinstance(rm, float)
        return rm

    @cached_property
    def h(self):
        "height of blade (m)"
        return self.rt-self.rh

    @cached_property
    def h_disk(self):
        "disk height of blade row (m)"
        xi = self.metal_angles.xi[0] if self.is_rotating else self.metal_angles.xi[-1]
        return np.abs(self.c*np.cos(xi) * 1.25)

    @cached_property
    def c(self):
        "chord length (m)"
        return self.h/self.AR

    @cached_property
    def tb(self):
        "blade max thickness (m)"
        return self.tbc * self.c

    @cached_property
    def Z(self):
        "number of blades in row (dimensionless)"
        Z = np.ceil(2*np.pi*self.rm/(self.sc*self.c))
        if not self.is_rotating and not Z % 2 == 0:
            Z -= 1
        return int(Z)

    @cached_property
    def s(self):
        "spacing between blades (m)"
        return 2*np.pi*self.rh/self.Z

    @cached_property
    def sh(self):
        "spacing to height (dimensionless)"
        return self.s/self.h

    @cached_property
    def sigma(self):
        "spacing between blades (dimensionless)"
        return 1 / self.sc

    @cached_property
    def deHaller(self):
        "deHaller factor (dimensionless)"
        return self.next_flow_station.W / self.flow_station.W

    @cached_property
    def Re(self):
        "Reynold's number of blade chord (dimensionless)"
        return self.stage_flow_station.rho * self.stage_flow_station.Vm * (self.c / self.stage_flow_station.mu)

    @cached_property
    def airfoil_type(self):
        # if self.stage_flow_station.MN < 0.7:
        #     return AirfoilType.NACA65
        # elif self.stage_flow_station.MN >= 0.7 and self.stage_flow_station.MN <= 1.20:
        #     return AirfoilType.DCA
        # raise ValueError("MN > 1.20 not currently supported")

        # TODO: only have support of DCA airfoil generation at the moment
        return AirfoilType.DCA

    @cached_property
    def metal_angles(self):
        if self.metal_angle_method == "JohnsenBullock":
            # beta1_rm: float = np.median(self.beta1)  # type: ignore
            # beta2_rm: float = np.median(self.beta2)  # type: ignore
            method_angle_method = JohnsenBullockMetalAngleMethod(self.beta1, self.beta2, self.sigma, self.tbc, self.airfoil_type)
            metal_angle_offset = method_angle_method.get_metal_angle_offset(self.deviation_iterations)
            return MetalAngles(self.beta1, self.beta2, metal_angle_offset.i, metal_angle_offset.delta)
        return MetalAngles(self.beta1, self.beta2, 0, 0)

    @cached_property
    def radii(self):
        "blade radii (m)"
        return np.linspace(self.rh, self.rt, self.N_stream, endpoint=True)

    @cached_property
    def flow_station(self):
        "flow station (FlowStation)"
        return self.stage_flow_station.copyStream(
            alpha=self.vortex.alpha(self.radii, self.is_rotating),
            radius=self.radii
        )

    @cached_property
    def next_flow_station(self):
        "next flow station (FlowStation)"
        assert self.next_stage_flow_station is not None
        return self.next_stage_flow_station.copyStream(
            alpha=self.vortex.alpha(self.radii, self.is_rotating),
            radius=self.radii
        )

    @cached_property
    def beta1(self):
        "blade inlet flow angle (rad)"
        if self.is_rotating:
            return self.flow_station.beta   # beta1
        return self.flow_station.alpha      # alpha2

    @cached_property
    def beta2(self):
        "blade outlet flow angle (rad)"
        if self.is_rotating:
            assert self.next_stage_flow_station is not None
            return self.next_stage_flow_station.beta                      # beta2

        assert self.next_stage_flow_station is not None or self.vortex.Rm == 0.5, "next_flow_station needs to be defined or Rc=0.5"
        if self.next_stage_flow_station is not None:
            return self.next_stage_flow_station.alpha                                    # alpha3
        return self.vortex.alpha(self.radii, is_rotating=not self.is_rotating)           # alpha3

    @cached_property
    def DF(self):
        "diffusion factor (dimensionless)"
        return 1-(np.cos(self.beta1)/np.cos(self.beta2))+(np.cos(self.beta1)/2)*self.sigma*(np.tan(self.beta1)-np.tan(self.beta2))

    @cached_property
    def airfoils(self):
        r0 = self.tb * 0.15
        # TODO: optimize this with Numba
        return [
            DCAAirfoil(self.c, self.metal_angles.theta[i], r0, self.tb, self.metal_angles.xi[i])
            for i in range(self.N_stream)
        ]

    @cached_property
    def attachment(self):
        max_length = 0.75*self.s if self.is_rotating else 1*self.s
        attachment = FirtreeAttachment(
            gamma=np.radians(40),
            beta=np.radians(40),
            ll=0.15*self.s,
            lu=0.2*self.s,
            Ri=0.05*self.s,
            Ro=0.025*self.s,
            R_dove=0.05*self.s,
            max_length=max_length,
            num_stages=2,
            disk_radius=self.rh,
            tolerance=0.0006,  # m, 0.5 mm
            include_top_arc=self.is_rotating
        )
        return attachment

    def to_cad_export(self):
        return BladeRowCadExport(
            stage_number=self.stage_number,
            disk_height=self.h_disk * MM,
            hub_radius=self.rh * MM,
            tip_radius=self.rt * MM,
            radii=self.radii * MM,
            airfoils=np.array([airfoil.get_coords() for airfoil in self.airfoils]) * MM,
            attachment=self.attachment.coords * MM,
            attachment_with_tolerance=self.attachment.coords_with_tolerance * MM,
            attachment_height=self.attachment.height * MM,
            attachment_bottom_width=self.attachment.bottom_width * MM,
            number_of_blades=self.Z,
            twist_angle=np.degrees(self.metal_angles.xi[-1]-self.metal_angles.xi[0]),
            is_rotating=self.is_rotating,
        )



from dataclasses import dataclass, field
from functools import cached_property
from typing import Union
import numpy as np

@dataclass
class Vortex:
    Um: float
    "mean blade velocity (m/s)"

    Vm: float
    "meridional flow velocity (m/s)"

    Rm: float
    "mean reaction rate (dimensionless)"

    psi_m: float
    "mean loading coefficient (dimensionless)"

    rm: float
    "mean radius (m)"

    def __post_init__(self):
        self.phi_m = self.Vm/self.Um

    def ctheta(self, r: Union[float, np.ndarray], is_rotating: bool):
        "absolute tangential velocity (m/s)"
        return np.nan

    def alpha(self, r: Union[float, np.ndarray], is_rotating: bool):
        "absolute flow angle (rad)"
        return np.arctan(self.ctheta(r, is_rotating)/self.Vm)


from dataclasses import dataclass, field
from typing import Union
import numpy as np
from turbodesigner.units import DEG

@dataclass
class MetalAngles:

    beta1: Union[float, np.ndarray]
    "blade inlet flow angle (rad)"

    beta2: Union[float, np.ndarray]
    "blade outlet flow angle (rad)"

    i: Union[float, np.ndarray]
    "blade incidence (rad)"

    delta: Union[float, np.ndarray]
    "blade deviation (rad)"

    kappa1: np.ndarray = field(init=False) 
    "inlet metal angle (rad)"

    kappa2: np.ndarray = field(init=False) 
    "outlet metal angle (rad)"

    theta: np.ndarray = field(init=False) 
    "camber angle (rad)"

    xi: np.ndarray = field(init=False) 
    "stagger angle (rad)"

    def __post_init__(self):
        self.kappa1 = np.asarray(self.beta1 - self.i)
        self.kappa2 = np.asarray(self.beta2 - self.delta)
        self.theta = np.asarray(self.kappa1-self.kappa2)
        self.xi = np.asarray((self.kappa1 + self.kappa2)/2)


"""
TurboDesigner CAD Module - Contains CAD functionality
"""

from turbodesigner.cad.blade import *
from turbodesigner.cad.casing import *
from turbodesigner.cad.common import *
from turbodesigner.cad.shaft import *

__all__ = [
    "blade",
    "casing",
    "common",
    "shaft",
]


"""
TurboDesigner Vortex Module - Contains vortex design functionality
"""

from turbodesigner.blade.vortex.common import *
from turbodesigner.blade.vortex.free_vortex import *

__all__ = [
    "common",
    "free_vortex",
]


from dataclasses import dataclass, field
from functools import cached_property
from typing import Optional
import cadquery as cq
from turbodesigner.cad.common import ExtendedWorkplane, FastenerPredicter
from turbodesigner.cad.blade import BladeCadModel, BladeCadModelSpecification
from turbodesigner.stage import StageCadExport
from turbodesigner.turbomachinery import TurbomachineryCadExport


@dataclass
class ShaftCadModelSpecification:
    is_simple: bool = False
    "whether to include attachment or fuse into shaft (bool)"

    stage_connect_length_to_heatset_thickness: float = 2.00
    "shaft stage connect to hub radius (dimensionless)"

    stage_connect_height_to_screw_head_diameter: float = 1.75
    "shaft stage connect to disk height (dimensionless)"

    stage_connect_padding_to_attachment_height: float = 1.25
    "shaft stage connect padding to attachment height (dimensionless)"

    stage_connect_heatset_diameter_to_disk_height: float = 0.05
    "shaft stage connect heatset diameter to disk height (dimensionless)"

    stage_connect_screw_quantity: int = 4
    "shaft stage connect screw quantity (dimensionless)"

    stage_connect_clearance: float = 0.5
    "shaft stage connect circular clearance (mm)"


@dataclass
class ShaftCadModel:
    stage: StageCadExport
    "turbomachinery stage"

    next_stage: Optional[StageCadExport] = None
    "next turbomachinery stage"

    spec: ShaftCadModelSpecification = field(default_factory=ShaftCadModelSpecification)
    "shaft cad model specification"

    def __post_init__(self):
        self.stage_connect_heatset = FastenerPredicter.predict_heatset(
            target_diameter=self.stage.rotor.disk_height*self.spec.stage_connect_heatset_diameter_to_disk_height,
        )
        self.stage_connect_length = self.stage_connect_heatset.nut_thickness * self.spec.stage_connect_length_to_heatset_thickness

        self.blade_cad_model = BladeCadModel(
            self.stage.rotor,
            spec=BladeCadModelSpecification(
                not self.spec.is_simple,
                screw_length_padding=self.stage_connect_length
            )
        )

        # TODO: Just used to get head diameter, length doesn't matter
        self.stage_connect_screw = FastenerPredicter.predict_screw(target_diameter=self.stage_connect_heatset.thread_diameter)

        self.stage_connect_height = self.stage_connect_screw.head_diameter * self.spec.stage_connect_height_to_screw_head_diameter
        stage_connect_padding = self.stage.rotor.attachment_height * self.spec.stage_connect_padding_to_attachment_height
        self.stage_connect_outer_radius = self.stage.rotor.hub_radius-stage_connect_padding
        self.stage_connect_inner_radius = self.stage_connect_outer_radius-self.stage_connect_length

        if self.next_stage:
            self.next_stage_shaft_cad_model = ShaftCadModel(self.next_stage, spec=self.spec)
            self.next_stage_stage_connect_screw = FastenerPredicter.predict_screw(
                target_diameter=self.next_stage_shaft_cad_model.stage_connect_heatset.thread_diameter,
                target_length=self.next_stage_shaft_cad_model.stage_connect_heatset.nut_thickness + (self.stage.stator.hub_radius - self.next_stage_shaft_cad_model.stage_connect_outer_radius)
            )

    @cached_property
    def shaft_stage_sector(self):
        sector_angle = 360 / self.stage.rotor.number_of_blades
        sector_cut_profile = (
            cq.Workplane('XZ')
            .transformed(rotate=(0, sector_angle/2, 0))
            .rect(self.stage.stator.hub_radius, self.stage.stage_height*2, centered=False)
            .revolve(sector_angle*(self.stage.rotor.number_of_blades-1), (0, 0, 0), (0, 1, 0))
        )

        shaft_profile = self.shaft_stage_assembly.objects["Stage Shaft"].obj
        assert shaft_profile is not None and isinstance(shaft_profile, cq.Workplane)
        shaft_sector_profile = (
            shaft_profile
            .cut(sector_cut_profile)
        )

        return shaft_sector_profile

    @cached_property
    def shaft_stage_assembly(self):
        base_assembly = cq.Assembly()
        blade_assembly = cq.Assembly()
        fastener_assembly = cq.Assembly()

        shaft_profile = (
            ExtendedWorkplane("XY")
            # Stator Disk
            .circle(self.stage.stator.hub_radius)
            .extrude(self.stage.stator.disk_height+self.stage.stage_gap)

            # Row Gap Transition Disk
            .faces(">Z")
            .workplane()
            .truncated_cone(
                start_radius=self.stage.stator.hub_radius,
                end_radius=self.stage.rotor.hub_radius,
                height=self.stage.row_gap
            )

            # Rotor Disk
            .faces(">Z")
            .workplane()
            .circle(self.stage.rotor.hub_radius)
            .extrude(self.stage.rotor.disk_height)
        )

        if not self.spec.is_simple:
            shaft_profile = (
                shaft_profile

                # Cut Attachments - TODO: make this operation faster
                .faces(">Z")
                .workplane()
                .polarArray(1.0001*self.stage.rotor.hub_radius, 0, 360, self.stage.rotor.number_of_blades)
                .eachpoint(
                    lambda loc: (
                        cq.Workplane("XY")
                        .polyline(self.stage.rotor.attachment_with_tolerance)  # type: ignore
                        .close()
                        .rotate((0, 0, 0), (0, 0, 1), 270)
                    ).val().located(loc), True)  # type: ignore
                .cutBlind(-self.stage.rotor.disk_height)

                # Shaft Male Connect
                .faces(">Z")
                .workplane()
                .circle(self.stage_connect_outer_radius)
                .extrude(self.stage_connect_height)

                # Shaft Connect Hole
                .faces(">Z")
                .workplane()
                .circle(self.stage_connect_inner_radius*1.001)
                .cutThruAll()

                # Blade Lock Screws
                .faces(">Z")
                .workplane(offset=-self.stage_connect_height-self.blade_cad_model.lock_screw.head_diameter*1.5)
                .polarArray(self.stage_connect_inner_radius, 0, 360, self.stage.rotor.number_of_blades)
                .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, 0), cq.Vector(0, 1, 0), -90))
                .clearanceHole(self.blade_cad_model.lock_screw, fit="Loose", baseAssembly=fastener_assembly)

                # Shaft Connect Heatsets
                .faces(">Z")
                .workplane(offset=-self.stage_connect_height/2)
                .polarArray(self.stage_connect_outer_radius, 0, 360, self.spec.stage_connect_screw_quantity)
                .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, 0), cq.Vector(0, 1, 0), 90))
                .insertHole(self.stage_connect_heatset, fit="Loose", baseAssembly=fastener_assembly, depth=self.stage_connect_heatset.nut_thickness)
            )
            if self.next_stage_shaft_cad_model:
                shaft_profile = (
                    # Next Shaft Female Connect
                    shaft_profile
                    .faces("<Z")
                    .workplane()
                    .circle(self.next_stage_shaft_cad_model.stage_connect_outer_radius + self.spec.stage_connect_clearance)
                    .cutBlind(-self.next_stage_shaft_cad_model.stage_connect_height)

                    # Next Shaft Connect Screws
                    .faces("<Z")
                    .workplane(offset=-self.next_stage_shaft_cad_model.stage_connect_height/2)
                    .polarArray(self.next_stage_shaft_cad_model.stage.rotor.hub_radius, 0, 360, self.spec.stage_connect_screw_quantity)
                    .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, 0), cq.Vector(0, 1, 0), 90))
                    .clearanceHole(self.next_stage_stage_connect_screw, fit="Loose", baseAssembly=fastener_assembly)
                )

        blade_vertical_offset = self.stage.stage_gap+self.stage.stator.disk_height+self.stage.row_gap+self.stage.rotor.disk_height/2
        blade_assembly_locs = (
            ExtendedWorkplane("XY")
            .polarArray(self.stage.rotor.hub_radius, 0, 360, self.stage.rotor.number_of_blades)
            .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, blade_vertical_offset), cq.Vector(0, 1, 0), 90))
            .vals()
        )

        for (i, blade_assembly_loc) in enumerate(blade_assembly_locs):
            assert isinstance(blade_assembly_loc, cq.Location)
            blade_assembly.add(self.blade_cad_model.blade_assembly, loc=blade_assembly_loc, name=f"Blade {i+1}")

        base_assembly.add(shaft_profile, name=f"Stage Shaft")
        base_assembly.add(blade_assembly, name="Blades")
        base_assembly.add(fastener_assembly, name="Fasteners")
        return base_assembly

    @staticmethod
    def shaft_assembly(turbomachinery: TurbomachineryCadExport, spec: ShaftCadModelSpecification = ShaftCadModelSpecification()):
        assembly = cq.Assembly()
        stage_height_offset = 0

        for i in range(0, len(turbomachinery.stages)):
            current_stage = turbomachinery.stages[i]
            next_stage = turbomachinery.stages[i+1] if i+1 < len(turbomachinery.stages) else current_stage

            stage_height_offset -= current_stage.stage_height + current_stage.stage_gap + current_stage.row_gap
            shaft_cad_model = ShaftCadModel(current_stage, next_stage, spec)
            assembly.add(shaft_cad_model.shaft_stage_assembly, loc=cq.Location(cq.Vector(0, 0, stage_height_offset)), name=f"Stage {current_stage.stage_number}")

        return assembly



from dataclasses import dataclass, field
from functools import cached_property
import cadquery as cq
import numpy as np
from turbodesigner.blade.row import BladeRowCadExport
from turbodesigner.cad.common import ExtendedWorkplane, FastenerPredicter


@dataclass
class BladeCadModelSpecification:
    include_attachment: bool = True
    "whether to include attachment (bool)"

    screw_length_padding: float = 0.00
    "screw length padding (dimensionless)"

    fastener_diameter_to_attachment_bottom_width: float = 0.25
    "blade attachment fastener to disk height (dimensionless)"


@dataclass
class BladeCadModel:
    blade_row: BladeRowCadExport
    "blade row"

    spec: BladeCadModelSpecification = field(default_factory=BladeCadModelSpecification)
    "blade cad model specification"

    @cached_property
    def lock_screw(self):
        return FastenerPredicter.predict_screw(
            target_diameter=self.heatset.thread_diameter,
            target_length=self.spec.screw_length_padding+self.heatset.nut_thickness
        )

    @cached_property
    def heatset(self):
        return FastenerPredicter.predict_heatset(
            target_diameter=self.spec.fastener_diameter_to_attachment_bottom_width*self.blade_row.attachment_bottom_width,
            max_thickness=self.blade_row.attachment_height*0.75
        )

    @cached_property
    def blade_assembly(self):
        base_assembly = cq.Assembly()
        fastener_assembly = cq.Assembly()

        start_airfoil = self.blade_row.airfoils[0]
        airfoil_vertical_offset = np.array([
            (np.max(start_airfoil[:, 0]) + np.min(start_airfoil[:, 0]))/2,
            (np.max(start_airfoil[:, 1]) + np.min(start_airfoil[:, 1]))/2
        ])

        # Hub Airfoil
        blade_profile = (
            cq.Workplane("XY")
            .polyline(start_airfoil - airfoil_vertical_offset)
            .close()
        )

        # Add all airfoil stations
        for i in range(0, len(self.blade_row.radii) - 1):
            blade_profile = (
                blade_profile
                .transformed(offset=cq.Vector(0, 0, self.blade_row.radii[i+1]-self.blade_row.radii[i]))
                .polyline(self.blade_row.airfoils[i+1] - airfoil_vertical_offset)
                .close()
            )

        if self.blade_row.is_rotating:
            hub_height_offset = self.blade_row.hub_radius*np.cos((2*np.pi / self.blade_row.number_of_blades) / 2)-self.blade_row.hub_radius
        else:
            hub_height_offset = 0

        blade_height = self.blade_row.tip_radius-self.blade_row.hub_radius
        path = (
            cq.Workplane("XZ")
            .lineTo(hub_height_offset, blade_height)
        )

        attachment_workplane = ExtendedWorkplane("YZ")
        if not self.blade_row.is_rotating:
            attachment_workplane = (
                attachment_workplane
                .transformed(rotate=(0, 0, 180), offset=(0, blade_height, 0))
            )
        # Attachment Profile
        attachment_profile = (
            attachment_workplane
            .polyline(self.blade_row.attachment)  # type: ignore
            .close()
            .extrude(self.blade_row.disk_height*0.5, both=True)

            .faces("<Z" if self.blade_row.is_rotating else ">Z")
            .workplane()
            .insertHole(self.heatset, depth=self.heatset.nut_thickness*0.9, baseAssembly=fastener_assembly)
        )

        blade_profile = (
            blade_profile
            .sweep(path, multisection=True, makeSolid=True)
        )

        if self.spec.include_attachment:
            blade_profile = (
                blade_profile
                .add(attachment_profile)
            )

        base_assembly.add(blade_profile, name="Blade")
        base_assembly.add(fastener_assembly, name="Fasteners")

        return base_assembly



from functools import cached_property
from dataclasses import dataclass
from typing import Optional, Union
import numpy as np

PROP_NON_STREAM_ERROR = "Property not allowed with streams"


class FluidConstants:
    MU_REF = 1.73E-5
    "reference dynamic viscocity at sea level ((N*s)/m**2)"

    PT_REF = 101325
    "reference pressure at sea level (Pa)"

    T0_REF = 288.15
    "reference temperature at sea level (K)"

    C = 110.4
    "sutherland constant (K)"


@dataclass
class FlowStation:
    "calculates flow station for an ideal gas"

    gamma: float = np.nan
    "ratio of specific heats (dimensionless)"

    Rs: float = np.nan
    "specific gas constant (J/(kg*K))"

    Tt: float = np.nan
    "total temperature (K)"

    Pt: float = np.nan
    "total pressure (Pa)"

    Vm: float = np.nan
    "meridional flow velocity (m/s)"

    mdot: float = np.nan
    "mass flow rate (kg/s)"

    B: float = 0.0
    "blockage factor (dimensionless)"

    alpha: Union[float, np.ndarray] = np.nan
    "absolute flow angle (rad)"

    N: float = np.nan
    "rotational speed (rpm)"

    radius: Union[float, np.ndarray] = np.nan
    "flow radius (m)"

    mixture: str = "Air"
    "mixture"

    is_stream: bool = False
    "whether station is 1D stream"

    def copyFlow(
        self,
        Tt: Optional[float] = None,
        Pt: Optional[float] = None,
        Vm: Optional[float] = None,
        mdot: Optional[float] = None,
        alpha: Optional[Union[float, np.ndarray]] = None,
        radius: Optional[Union[float, np.ndarray]] = None,
    ):
        "copies all elements of FlowStation (FlowStation)"
        return FlowStation(
            gamma=self.gamma,
            Rs=self.Rs,
            Tt=self.Tt if Tt is None else Tt,
            Pt=self.Pt if Pt is None else Pt,
            Vm=self.Vm if Vm is None else Vm,
            mdot=self.mdot if mdot is None else mdot,
            B=self.B,
            alpha=self.alpha if alpha is None else alpha,
            N=self.N,
            radius=self.radius if radius is None else radius,
            mixture=self.mixture
        )

    def copyStream(
        self,
        alpha: Optional[Union[float, np.ndarray]] = None,
        radius: Optional[Union[float, np.ndarray]] = None,
    ):
        """copies stream elements of FlowStation (FlowStation)

           excludes:
                * mdot - mass flow rate

        """
        return FlowStation(
            gamma=self.gamma,
            Rs=self.Rs,
            Tt=self.Tt,
            Pt=self.Pt,
            Vm=self.Vm,
            alpha=self.alpha if alpha is None else alpha,
            N=self.N,
            radius=self.radius if radius is None else radius,
            is_stream=True,
            mixture=self.mixture
        )

    @cached_property
    def h(self):
        "static enthalpy (J/kg*K)"
        return self.T*self.Cp

    @cached_property
    def ht(self):
        "total enthalpy (J/kg*K)"
        return self.h + (self.V**2)/2

    @cached_property
    def Cp(self):
        "specific heat at constant pressure (J/(kg*K))"
        return self.Rs*self.gamma/(self.gamma - 1)

    @cached_property
    def T(self):
        "static temperature (K)"
        return self.Tt - (self.V**2)/(2*self.Cp)

    @cached_property
    def Ttr(self):
        "total realtive temperature (K)"
        return self.Tt + (self.W**2 - self.V**2)/(2*self.Cp)

    @cached_property
    def P(self):
        "static pressure (Pa)"
        return self.Pt*(self.T/self.Tt)**(self.gamma/(self.gamma - 1))

    @cached_property
    def Ptr(self):
        "total relative pressure (Pa)"
        return self.Pt*(self.Ttr/self.Tt)**(self.gamma/(self.gamma - 1))

    @cached_property
    def rho(self):
        "density (kg/m**3)"
        return self.P/(self.T*self.Rs)

    @cached_property
    def q(self):
        "dynamic pressure (Pa)"
        return 0.5*self.rho*self.Vm**2

    @cached_property
    def a(self):
        "speed of sound in medium (m/s)"
        return np.sqrt(self.T*self.Rs*self.gamma)

    @cached_property
    def mu(self):
        "dynamic velocity using Sutherland's formula ((N*s)/m**2)"
        return FluidConstants.MU_REF * ((self.T / FluidConstants.T0_REF)**1.5) * ((FluidConstants.T0_REF + FluidConstants.C) / (self.T + FluidConstants.C))

    @cached_property
    def MN(self):
        "mach number (dimensionless)"
        return self.Vm/self.a

    @cached_property
    def Vcr(self):
        "critical velocity (m/s)"
        return np.sqrt(((2*self.gamma)/(self.gamma+1)) * self.Rs*self.Tt)

    @cached_property
    def U(self):
        "blade velocity (m/s)"
        return FlowStation.calc_U(self.N, self.radius)

    @cached_property
    def omega(self):
        "blade angular velocity (rad/s)"
        return self.U/self.radius

    @cached_property
    def Vtheta(self):
        "absolute tangential velocity (m/s)"
        return self.Vm*np.tan(self.alpha)

    @cached_property
    def V(self):
        "absolute flow velocity (m/s)"
        if np.isnan(self.alpha).all():
            return self.Vm
        return self.Vm/np.cos(self.alpha)

    @cached_property
    def Wtheta(self):
        "relative tangential flow velocity (m/s)"
        return self.Vtheta - self.U

    @cached_property
    def beta(self):
        "relative flow angle (rad)"
        return np.arctan(self.Wtheta/self.Vm)

    @cached_property
    def W(self):
        "relative flow velocity (m/s)"
        return self.Vm/np.cos(self.beta)

    # %% Annular Properties
    @cached_property
    def A_flow(self):
        "cross-sectional flow area (m**2)"
        assert not self.is_stream, PROP_NON_STREAM_ERROR
        return self.mdot/(self.rho*self.Vm)

    @cached_property
    def A_phys(self):
        "physical cross sectional area (m**2)"
        return self.A_flow*(self.B + 1)

    @cached_property
    def outer_radius(self):
        "flow outer radius (m)"
        assert not self.is_stream, PROP_NON_STREAM_ERROR
        return self.A_phys/(4*np.pi*self.radius) + self.radius

    @cached_property
    def inner_radius(self):
        "flow inner radius (m)"
        assert not self.is_stream, PROP_NON_STREAM_ERROR
        return 2*self.radius - self.outer_radius

    @staticmethod
    def calc_radius_from_ht(ht: float, A_phys: Union[float, np.ndarray]):
        """calculates radius from hub to tip ratio

        Parameters
        ==========

        ht: float
            hub to tip ratio (dimensionless)

        A_phys: float
            physical cross sectional area (m**2)

        """

        outer_radius = np.sqrt(A_phys / (np.pi*(1-ht**2)))
        inner_radius = ht * outer_radius
        return (outer_radius + inner_radius) / 2

    @staticmethod
    def calc_U(N: float, radius: Union[float, np.ndarray]):
        """calculates blade velocity

        Parameters
        ==========

        N: float
            rotational speed (rpm)

        radius: float
            blade radius (m)

        """

        return 2*np.pi*N*radius/60

    def set_radius(self, ht: float):
        """sets radius from hub to tip ratio

        Parameters
        ==========

        ht: float
            hub to tip ratio (dimensionless)
        """
        self.radius = FlowStation.calc_radius_from_ht(ht, self.A_phys)



from typing import Callable, List, Optional
import cadquery as cq
import cq_warehouse.extensions as cq_warehouse_extensions
from cq_warehouse.fastener import SocketHeadCapScrew, HeatSetNut
import numpy as np
import re


class ExtendedWorkplane(cq.Workplane):
    clearanceHole = cq_warehouse_extensions._clearanceHole
    tapHole = cq_warehouse_extensions._tapHole
    threadedHole = cq_warehouse_extensions._threadedHole
    insertHole = cq_warehouse_extensions._insertHole
    pressFitHole = cq_warehouse_extensions._pressFitHole

    def ring(self, radius: float, thickness: float, depth: float) -> "ExtendedWorkplane":
        return (
            self
            .circle(radius)
            .circle(radius - thickness)
            .extrude(depth)
        )

    def truncated_cone(self, start_radius: float, end_radius: float, height: float):
        plane_world_coords = self.plane.toWorldCoords((0, 0, 0))
        path = cq.Workplane("XZ").moveTo(0, plane_world_coords.z).lineTo(0, height + plane_world_coords.z)

        return (
            self
            .circle(start_radius)
            .transformed(offset=cq.Vector(0, 0, height))
            .circle(end_radius)
            .sweep(path, multisection=True, makeSolid=True)
            .clean()
        )

    def hollow_truncated_cone(self, inner_start_radius: float, inner_end_radius: float, height: float, start_thickness: float, end_thickness: float):
        outer_radius = inner_start_radius + start_thickness
        return (
            self
            .truncated_cone(outer_radius, outer_radius, height)
            .cut(
                ExtendedWorkplane("XY")
                .truncated_cone(inner_start_radius, inner_end_radius, height)
            )
        )

    def mutatePoints(self, callback: Callable[[cq.Location], cq.Location]):
        for (i, loc) in enumerate(self.objects):
            if isinstance(loc, cq.Vertex) or isinstance(loc, cq.Vector):
                loc = cq.Location(self.plane, loc.toTuple())

            assert isinstance(loc, cq.Location)
            mutated_loc = callback(loc)
            self.objects[i] = mutated_loc

        return self
 

class FastenerPredicter:
    @staticmethod
    def get_nominal_size(
        target_diameter: float,
        nominal_size_range: List[str],
    ):
        for nominal_size in nominal_size_range:
            nominal_diameter = float(nominal_size.split("-")[0].replace("M", ""))
            if target_diameter <= nominal_diameter:
                return nominal_size

        raise ValueError(f"nominal size for target diameter {target_diameter} could not be found")

    @staticmethod
    def get_nominal_length(
        target_length: float,
        nominal_length_range: List[float],
    ):
        for nominal_length in nominal_length_range:
            if target_length <= nominal_length:
                return nominal_length
        raise ValueError(f"nominal length for target length {target_length} could not be found")


    @staticmethod
    def predict_heatset(target_diameter: float, max_thickness: Optional[float] = None, type: str = "Hilitchi"):
        nominal_size_range = HeatSetNut.sizes(type)
        last_acceptable_height_heatset: Optional[HeatSetNut] = None
        for nominal_size in nominal_size_range:
            heatset = HeatSetNut(
                size=nominal_size,
                fastener_type=type,
                simple=True,
            )
            if max_thickness and heatset.nut_thickness <= max_thickness:
                last_acceptable_height_heatset = heatset
            if target_diameter <= heatset.nut_diameter :
                if max_thickness and heatset.nut_thickness > max_thickness:
                    assert last_acceptable_height_heatset is not None, f"no heasets are valid for max height {max_thickness}, closest heatset {heatset.nut_diameter}"
                    return last_acceptable_height_heatset
                return heatset
        raise ValueError(f"nominal size for target diameter {target_diameter} could not be found")


    @staticmethod
    def predict_screw(target_diameter: float, target_length: Optional[float] = None, type: str = "iso4762"):
        nominal_length_range = SocketHeadCapScrew.nominal_length_range[type]
        predicted_length = nominal_length_range[0]
        if target_length:
            predicted_length = FastenerPredicter.get_nominal_length(target_length, nominal_length_range)
        
        nominal_size_range = SocketHeadCapScrew.sizes(type)
        predicted_size = FastenerPredicter.get_nominal_size(target_diameter, nominal_size_range)

        return SocketHeadCapScrew(predicted_size, predicted_length, type)



import os
from setuptools import setup, find_packages

# Read the contents of README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Read requirements.txt
with open("requirements.txt", "r", encoding="utf-8") as f:
    requirements = []
    for line in f:
        line = line.strip()
        # Skip comments and empty lines
        if not line or line.startswith("#"):
            continue
        # Skip git+ dependencies for PyPI compatibility
        if line.startswith("git+"):
            # Git dependencies aren't directly supported by PyPI
            # They'll be handled through extras_require
            continue

        requirements.append(line)

setup(
    name="turbodesigner",
    version="1.0.0",
    author="Open Orion, Inc.",
    description="An open-source turbomachinery designer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/OpenOrion/turbodesigner",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
    python_requires=">=3.8",
    install_requires=requirements,
    include_package_data=True,
    package_data={
        "turbodesigner": ["**/*.json"],
    },
)


from typing import Callable, Optional
import numpy as np
import pandas as pd
from turbodesigner.stage import Stage
from turbodesigner.turbomachinery import Turbomachinery
from turbodesigner.units import DEG, BAR


def get_hub_tip_dict_from_export(
    stage: Stage,
    table_dict: dict[str, dict],
    export_dict: dict,
    group_name: Optional[str] = None
):
    for (key, value) in export_dict.items():
        group_name = group_name or f"Stage {stage.stage_number}"
        table_dict["Hub"][(group_name, key)] = value[0] if isinstance(value, np.ndarray) else value
        table_dict["Mean"][(group_name, key)] = np.median(value) if isinstance(value, np.ndarray) else value
        table_dict["Tip"][(group_name, key)] = value[-1] if isinstance(value, np.ndarray) else value
    return table_dict


def get_hub_mean_tip_table(
    turbomachinery: Turbomachinery,
    to_export_dict: Callable[[Stage], dict],
    is_multi_row=False
):
    table = {
        "Hub": dict(),
        "Mean": dict(),
        "Tip": dict()
    }

    for stage in turbomachinery.stages:
        export_dict = to_export_dict(stage)
        if is_multi_row:
            table = get_hub_tip_dict_from_export(stage, table, export_dict["Rotor"], f"Stage {stage.stage_number} - Rotor")
            table = get_hub_tip_dict_from_export(stage, table, export_dict["Stator"], f"Stage {stage.stage_number} - Stator")
        else:
            table = get_hub_tip_dict_from_export(stage, table, export_dict)

    return pd.DataFrame(table)


def get_rotor_stator_table(turbomachinery: Turbomachinery, to_export_dict: Callable[[Stage], dict]):
    table_dict = dict()
    for stage in turbomachinery.stages:
        export_dict = to_export_dict(stage)
        for (key, value) in export_dict.items():
            if key not in table_dict:
                table_dict[key] = dict()
            table_dict[key][(f"Stage {stage.stage_number}", "Rotor")] = value["Rotor"]
            table_dict[key][(f"Stage {stage.stage_number}", "Stator")] = value["Stator"]
    return pd.DataFrame(table_dict)


class TurbomachineryExporter:
    @staticmethod
    def turbomachinery_properties(turbomachinery: Turbomachinery):
        properties = {
            "gamma (dimensionless)": turbomachinery.gamma,
            "cx (m/s)": turbomachinery.cx,
            "N (rpm)": turbomachinery.N,
            "Rs (J/(kgK))": turbomachinery.Rs,
            "mdot (kg/s)": turbomachinery.mdot,
            "PR (dimensionless)": turbomachinery.PR,
            "Pt1 (bar)": turbomachinery.Pt*BAR,
            "Tt1 (K)": turbomachinery.Tt,
            "eta_isen (dimensionless)": turbomachinery.eta_isen,
            "eta_poly (dimensionless)": turbomachinery.eta_poly,
            "N_stg": turbomachinery.N_stg,
            "B_in (dimensionless)":  turbomachinery.B_in,
            "B_out (dimensionless)": turbomachinery.B_out,
            "ht (dimensionless)": turbomachinery.ht,
        }
        return pd.DataFrame.from_dict(properties, orient='index')

    @staticmethod
    def stage_properties(turbomachinery: Turbomachinery):
        return pd.DataFrame(
            [
                {
                    "Stage": stage.stage_number,
                    "Delta_Tt (K)": stage.Delta_Tt,
                    "Delta_ht (J/kg)": stage.Delta_ht,
                    "PR (dimensionless)": stage.PR,
                    "R (dimensionless)": stage.R,
                    "phi (dimensionless)": stage.phi,
                    "psi (dimensionless)": stage.psi
                }
                for stage in turbomachinery.stages
            ],
        )

    @staticmethod
    def stage_fluid_properties(turbomachinery: Turbomachinery):
        return pd.DataFrame(
            [
                {
                    "Stage": stage.stage_number,
                    "Tt1 (K)": stage.inlet_flow_station.Tt,
                    "Pt1 (bar)": stage.inlet_flow_station.Pt * BAR,
                    "ht1 (J/kg*K)": stage.inlet_flow_station.ht,
                    "T1 (K)": stage.inlet_flow_station.T,
                    "P1 (bar)": stage.inlet_flow_station.P * BAR,
                    "H1 (K)": stage.inlet_flow_station.h,
                    "rho1 (kg/m^3)": stage.inlet_flow_station.rho,

                    "Tt2 (K)": stage.mid_flow_station.Tt,
                    "Pt2 (bar)": stage.mid_flow_station.Pt * BAR,
                    "ht2 (J/kg*K)": stage.mid_flow_station.ht,
                    "T2 (K)": stage.mid_flow_station.T,
                    "P2 (bar)": stage.mid_flow_station.P * BAR,
                    "H2 (K)": stage.mid_flow_station.h,
                    "rho2 (kg/m^3)": stage.mid_flow_station.rho,
                }
                for stage in turbomachinery.stages
            ]
        )

    @staticmethod
    def annulus(turbomachinery: Turbomachinery):
        return get_rotor_stator_table(
            turbomachinery,
            lambda stage: {
                "rh (m)": {
                    "Rotor": stage.rotor.rh,
                    "Stator": stage.stator.rh,
                },
                "rt (m)": {
                    "Rotor": stage.rotor.rt,
                    "Stator": stage.stator.rt,
                },
                "rm (m)": {
                    "Rotor": stage.rotor.rm,
                    "Stator": stage.stator.rm,
                }
            }
        )

    @staticmethod
    def velocity_triangle(turbomachinery: Turbomachinery):
        return get_hub_mean_tip_table(
            turbomachinery,
            lambda stage: {
                "Vm (m/s)": stage.rotor.flow_station.Vm,
                "U (m/s)": stage.rotor.flow_station.U,
                "Vθ1 (m/s)": stage.rotor.flow_station.Vtheta,
                "V1 (m/s)": stage.rotor.flow_station.V,
                "Wθ1 (m/s)": stage.rotor.flow_station.Wtheta,
                "W1 (m/s)": stage.rotor.flow_station.W,
                "beta1 (deg)": stage.rotor.flow_station.beta * DEG,
                "alpha1 (deg)": stage.rotor.flow_station.alpha * DEG,
                "Vθ2 (m/s)": stage.stator.flow_station.Vtheta,
                "V2 (m/s)": stage.stator.flow_station.V,
                "Wθ2 (m/s)": stage.stator.flow_station.Wtheta,
                "W2 (m/s)": stage.stator.flow_station.W,
                "beta2 (deg)": stage.stator.flow_station.beta * DEG,
                "alpha2 (deg)": stage.stator.flow_station.alpha * DEG,
            }
        )

    @staticmethod
    def blade_angles(turbomachinery: Turbomachinery):
        return get_hub_mean_tip_table(
            turbomachinery,
            lambda stage: {
                "Rotor": {
                    "delta (deg)": stage.rotor.metal_angles.delta * DEG,
                    "i (deg)": stage.rotor.metal_angles.i * DEG,
                    "kappa1 (deg)": stage.rotor.metal_angles.kappa1 * DEG,
                    "kappa2 (deg)": stage.rotor.metal_angles.kappa2 * DEG,
                    "theta (deg)": stage.rotor.metal_angles.theta * DEG,
                    "xi (deg)": stage.rotor.metal_angles.xi * DEG,
                },
                "Stator": {
                    "delta (deg)": stage.stator.metal_angles.delta * DEG,
                    "i (deg)": stage.stator.metal_angles.i * DEG,
                    "kappa1 (deg)": stage.stator.metal_angles.kappa1 * DEG,
                    "kappa2 (deg)": stage.stator.metal_angles.kappa2 * DEG,
                    "theta (deg)": stage.stator.metal_angles.theta * DEG,
                    "xi (deg)": stage.stator.metal_angles.xi * DEG,
                }
            },
            is_multi_row=True
        )

    @staticmethod
    def blade_properties(turbomachinery: Turbomachinery):
        return get_rotor_stator_table(
            turbomachinery,
            lambda stage: {
                "sc (dimensionless)": {
                    "Rotor": stage.rotor.sc,
                    "Stator": stage.stator.sc,
                },
                "AR (dimensionless)": {
                    "Rotor": stage.rotor.AR,
                    "Stator": stage.stator.AR,
                },
                "tbc (dimensionless)": {
                    "Rotor": stage.rotor.tbc,
                    "Stator": stage.stator.tbc,
                },
                "sigma (dimensionless)": {
                    "Rotor": stage.rotor.sigma,
                    "Stator": stage.stator.sigma,
                },
                "c (m)": {
                    "Rotor": stage.rotor.c,
                    "Stator": stage.stator.c,
                },
                "h (m)": {
                    "Rotor": stage.rotor.h,
                    "Stator": stage.stator.h,
                }
            }
        )



from typing import Union
import numpy as np
from turbodesigner.blade.vortex.common import Vortex

class FreeVortex(Vortex):
    def ctheta(self, r: Union[float, np.ndarray], is_rotating: bool):
        mu = r/self.rm
        a = self.Um*(1-self.Rm)
        b = (1/2)*self.psi_m*self.Um
        sign = -1 if is_rotating else 1
        return (a/mu) + sign*(b/mu)


"""
TurboDesigner Airfoils Module - Contains airfoil design functionality
"""

from turbodesigner.airfoils.common import *
from turbodesigner.airfoils.DCA import *

__all__ = [
    "common",
    "DCA",
]


"""
TurboDesigner Metal Angle Methods Module - Contains methods for calculating metal angles
"""

from turbodesigner.blade.metal_angle_methods.johnsen_bullock import *

__all__ = [
    "johnsen_bullock",
]


import numpy as np
from turbodesigner.turbomachinery import Turbomachinery
import plotly.graph_objects as go
from IPython.display import Image, display


class TurbomachineryVisualizer:
    @staticmethod
    def visualize_annulus(turbomachinery: Turbomachinery, is_interactive=False):
        rotors = [
            np.array([
                [(stage.stage_number-1.), stage.rotor.rt],
                [(stage.stage_number-1.)+0.5, stage.stator.rt],
                [(stage.stage_number-1.)+0.5, stage.stator.rh],
                [(stage.stage_number-1.), stage.rotor.rh],
                [(stage.stage_number-1.), stage.rotor.rt],
            ])
            for stage in turbomachinery.stages
        ]

        stators = [
            np.array([
                [(stage.stage_number-1.)+0.5, stage.stator.rt],
                [stage.stage_number, (stage.next_stage.rotor if stage.next_stage else stage.stator).rt],
                [stage.stage_number, (stage.next_stage.rotor if stage.next_stage else stage.stator).rh],
                [(stage.stage_number-1.)+0.5, stage.stator.rh],
                [(stage.stage_number-1.)+0.5, stage.stator.rt],
            ])
            for stage in turbomachinery.stages
        ]


        fig = go.Figure(
            layout=go.Layout(
                title=go.layout.Title(text="Annulus"),
            )
        )
        for (i, rotor) in enumerate(rotors):
            fig.add_trace(go.Scatter(
                x=rotor[:, 0],
                y=rotor[:, 1],
                fill="toself",
                fillcolor="red",
                line_color="red",
                legendgroup="rotor",
                legendgrouptitle_text="Rotor",
                name=f"Rotor {i+1}"
            ))

        for (i, stator) in enumerate(stators):
            fig.add_trace(go.Scatter(
                x=stator[:, 0],
                y=stator[:, 1],
                fill="toself",
                fillcolor="blue",
                line_color="blue",
                legendgroup="stator",
                legendgrouptitle_text="Stator",
                name=f"Stator {i+1}"
            ))

        

        if is_interactive:
            fig.show()
        else:
            image = Image(fig.to_image(format="png", width=800, height=500, scale=2))
            display(image)


from dataclasses import dataclass, field
from functools import cached_property
from typing import Optional
from turbodesigner.cad.blade import BladeCadModel, BladeCadModelSpecification
from turbodesigner.cad.common import ExtendedWorkplane, FastenerPredicter
from turbodesigner.turbomachinery import TurbomachineryCadExport
from turbodesigner.stage import StageCadExport
import cadquery as cq
import numpy as np


@dataclass
class CasingCadModelSpecifciation:
    is_simple: bool = False
    "create simple assembly (bool)"

    casing_thickness_to_inlet_radius: float = 0.25
    "casing thickness to tip radius of first stage (dimensionless)"

    stage_connect_height_to_screw_head_diameter: float = 1.75
    "casing stage connect to disk height (dimensionless)"

    stage_connect_padding_to_attachment_height: float = 1.25
    "casing stage connect padding to attachment height (dimensionless)"

    stage_connect_heatset_diameter_to_disk_height: float = 0.25
    "casing stage connect heatset diameter to disk height (dimensionless)"

    stage_connect_screw_quantity: int = 4
    "casing stage connect screw quantity (dimensionless)"

    half_connect_width_to_casing_thickness: float = 0.5
    "casing half connect width to casing thickness"


@dataclass
class CasingCadModel:
    stage: StageCadExport
    "turbomachinery stage"

    first_stage: StageCadExport
    "turbomachinery first stage"

    previous_stage: Optional[StageCadExport] = None
    "turbomachinery next stage"

    spec: CasingCadModelSpecifciation = field(default_factory=CasingCadModelSpecifciation)
    "casing cad model specification"

    def __post_init__(self):
        self.stage_connect_heatset = FastenerPredicter.predict_heatset(
            target_diameter=self.stage.rotor.disk_height*self.spec.stage_connect_heatset_diameter_to_disk_height,
        )
        self.casing_thickness = self.spec.casing_thickness_to_inlet_radius * self.first_stage.rotor.tip_radius
        self.casing_radius = self.first_stage.rotor.tip_radius+self.casing_thickness
        stage_connect_padding = self.spec.stage_connect_padding_to_attachment_height * self.stage.stator.attachment_height
        self.stage_connect_length = self.casing_thickness - stage_connect_padding - self.stage.stator.attachment_height

        self.blade_cad_model = BladeCadModel(
            self.stage.stator,
            spec=BladeCadModelSpecification(
                not self.spec.is_simple,
                screw_length_padding=self.casing_thickness-self.stage.stator.attachment_height
            )
        )

        self.stage_connect_screw = FastenerPredicter.predict_screw(
            target_diameter=self.stage_connect_heatset.thread_diameter,
            target_length=self.stage_connect_length+self.stage_connect_heatset.nut_thickness
        )

        self.stage_connect_height = self.stage_connect_screw.head_diameter * self.spec.stage_connect_height_to_screw_head_diameter
        self.stage_connect_outer_radius = self.casing_radius
        self.stage_connect_inner_radius = self.stage_connect_outer_radius-self.stage_connect_length

        self.previous_stage_casing_cad_model: Optional[CasingCadModel] = None
        if self.previous_stage:
            self.previous_stage_casing_cad_model = CasingCadModel(self.previous_stage, self.first_stage, spec=self.spec)

        # Half Connector
        self.sector_angle = 360 / self.stage.stator.number_of_blades
        self.half_connect_thickness = self.casing_radius*np.sin(np.radians(self.sector_angle) / 2)
        self.half_connect_width = self.casing_thickness*self.spec.half_connect_width_to_casing_thickness
        self.half_connect_height = self.stage.stage_height + self.stage.row_gap
        self.half_connect_heatset = FastenerPredicter.predict_heatset(
            target_diameter=self.half_connect_width*0.5,
        )
        self.half_connect_screw = FastenerPredicter.predict_screw(
            target_diameter=self.half_connect_heatset.thread_diameter,
            target_length=self.half_connect_thickness*0.75+self.half_connect_heatset.nut_thickness
        )

    @cached_property
    def casing_stage_assembly(self):
        base_assembly = cq.Assembly()
        blade_assembly = cq.Assembly()
        fastener_assembly = cq.Assembly()

        casing_cut_profile = (
            # Stator Disk
            ExtendedWorkplane("XY")
            .transformed(offset=(0, 0, -self.stage_connect_height))
            .circle(self.stage.stator.tip_radius*1.001)
            .extrude(self.stage.stator.disk_height+self.stage.stage_gap+self.stage_connect_height)

            # Transition Disk
            .faces(">Z")
            .workplane()
            .truncated_cone(
                start_radius=self.stage.stator.tip_radius,
                end_radius=self.stage.rotor.tip_radius,
                height=self.stage.row_gap
            )

            # Rotor Disk
            .faces(">Z")
            .workplane()
            .circle(self.stage.rotor.tip_radius)
            .extrude(self.stage.rotor.disk_height)
        )

        casing_profile = (
            ExtendedWorkplane("XY")
            .circle(self.first_stage.rotor.tip_radius + self.casing_thickness)
            .extrude(self.half_connect_height)
        )

        if not self.spec.is_simple:
            casing_profile = (
                casing_profile

                # Stage Shaft Connect
                .faces("<Z")
                .workplane()
                .circle(self.stage_connect_outer_radius)
                .circle(self.stage_connect_inner_radius)
                .extrude(self.stage_connect_height)

                # Add Half Connect
                .add(
                    cq.Workplane("XY")
                    .transformed(rotate=(0, 0, 90), offset=(0, 0, -self.stage_connect_height))
                    .box(self.half_connect_thickness*2, (self.casing_radius+self.half_connect_width)*2, self.half_connect_height, centered=(True, True, False))
                )
            )

        casing_profile = (
            casing_profile
            .cut(casing_cut_profile)
        )

        if not self.spec.is_simple:
            casing_profile = (
                casing_profile
                # Cut Attachments - TODO: make this operation faster
                .faces("<Z")
                .workplane(offset=-self.stage_connect_height)
                .transformed(rotate=(0, 0, -self.sector_angle/2))
                .polarArray(self.stage.stator.tip_radius, 0, 360, self.stage.stator.number_of_blades)
                .eachpoint(
                    lambda loc: (
                        cq.Workplane("XY")
                        .polyline(self.stage.stator.attachment_with_tolerance)  # type: ignore
                        .close()
                        .rotate((0, 0, 0), (0, 0, 1), 90)

                    ).val().located(loc), True)  # type: ignore
                .cutBlind(-self.stage.stator.disk_height)

                # Blade Lock Screws
                .faces("<Z")
                .workplane(offset=-self.stage_connect_height-self.stage.stator.attachment_height)
                .transformed(rotate=(0, 0, -self.sector_angle/2))
                .polarArray(self.stage_connect_outer_radius, 0, 360, self.stage.stator.number_of_blades)
                .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, 0), cq.Vector(0, 1, 0), 90))
                .clearanceHole(self.blade_cad_model.lock_screw, depth=self.blade_cad_model.lock_screw.length, fit="Loose", baseAssembly=fastener_assembly)

                # Stage Shaft Connect
                .faces("<Z")
                .workplane()
                .circle(self.stage_connect_inner_radius)
                .cutBlind(-self.stage_connect_height)

                # Stage Shaft Connect Screws
                .faces("<Z")
                .workplane(offset=-self.stage_connect_height/2)
                .transformed(rotate=(0, 0, 45))
                .polarArray(self.stage_connect_outer_radius, 0, 360, self.spec.stage_connect_screw_quantity)
                .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, 0), cq.Vector(0, 1, 0), 90))
                .clearanceHole(self.stage_connect_screw, fit="Loose", baseAssembly=fastener_assembly)
            )
            if self.previous_stage_casing_cad_model:
                casing_profile = (
                    casing_profile
                    # Previous Stage Shaft Connect
                    .faces(">Z")
                    .workplane()
                    .circle(self.previous_stage_casing_cad_model.stage_connect_outer_radius)
                    .circle(self.previous_stage_casing_cad_model.stage_connect_inner_radius)
                    .cutBlind(-self.previous_stage_casing_cad_model.stage_connect_height)

                    # Previous Stage Shaft Connect Heatsets
                    .faces(">Z")
                    .workplane(offset=-self.previous_stage_casing_cad_model.stage_connect_height/2)
                    .transformed(rotate=(0, 0, 45))
                    .polarArray(self.previous_stage_casing_cad_model.stage_connect_inner_radius, 0, 360, self.spec.stage_connect_screw_quantity)
                    .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, 0), cq.Vector(0, 1, 0), 90))
                    .insertHole(self.previous_stage_casing_cad_model.stage_connect_heatset, fit="Loose", baseAssembly=fastener_assembly, depth=self.previous_stage_casing_cad_model.stage_connect_heatset.nut_thickness)
                )

        if not self.spec.is_simple:
            left_casing_profile = (
                casing_profile
                .transformed(rotate=(90, -45, 0))
                .split(keepBottom=True)
                .faces("<Z")
                .workplane()
                .transformed(rotate=(-90, 0, 0), offset=(0, 0, -self.half_connect_height/2))
                .rect(self.casing_radius*2+self.half_connect_width, self.half_connect_height*0.75, forConstruction=True)
                .vertices()
                .insertHole(self.half_connect_heatset, fit="Loose", baseAssembly=fastener_assembly, depth=self.half_connect_heatset.nut_thickness)
            )

            right_casing_profile = (
                casing_profile
                .transformed(rotate=(90, 135, 0))
                .split(keepBottom=True)
                .faces("<Z")
                .workplane()
                .transformed(rotate=(-90, 0, 180), offset=(0, self.half_connect_thickness, -self.half_connect_height/2))
                .rect(self.casing_radius*2+self.half_connect_width, self.half_connect_height*0.75, forConstruction=True)
                .vertices()
                .clearanceHole(self.half_connect_screw, depth=self.half_connect_screw.length, fit="Loose", baseAssembly=fastener_assembly)
            )

            base_assembly.add(left_casing_profile, name=f"Left Casing")
            base_assembly.add(right_casing_profile, name=f"Right Casing")
        else:
            base_assembly.add(casing_profile, name=f"Casing")

        blade_vertical_offset = self.stage.stator.disk_height/2
        blade_assembly_locs = (
            ExtendedWorkplane("XY")
            .polarArray(self.stage.stator.hub_radius, 0, 360, self.stage.stator.number_of_blades)
            .mutatePoints(lambda loc: loc * cq.Location(cq.Vector(0, 0, blade_vertical_offset), cq.Vector(0, 1, 0), 90))
            .vals()
        )

        for (i, blade_assembly_loc) in enumerate(blade_assembly_locs):
            assert isinstance(blade_assembly_loc, cq.Location)
            blade_assembly.add(self.blade_cad_model.blade_assembly, loc=blade_assembly_loc, name=f"Blade {i+1}")
        blade_assembly.rotate((0, 0, 1), -self.sector_angle/2)  # type: ignore

        base_assembly.add(blade_assembly, name="Blades")
        base_assembly.add(fastener_assembly, name="Fasteners")

        return base_assembly

    @staticmethod
    def casing_assembly(
        turbomachinery: TurbomachineryCadExport,
        spec: CasingCadModelSpecifciation = CasingCadModelSpecifciation()
    ):
        assembly = cq.Assembly()
        stage_height_offset = 0
        first_stage = turbomachinery.stages[0]
        previous_stage: Optional[StageCadExport] = None
        for i in range(0, len(turbomachinery.stages)):
            current_stage = turbomachinery.stages[i]

            stage_height_offset -= current_stage.stage_height + current_stage.stage_gap + current_stage.row_gap
            casing_cad_model = CasingCadModel(current_stage, first_stage, previous_stage, spec)
            assembly.add(casing_cad_model.casing_stage_assembly, loc=cq.Location(cq.Vector(0, 0, stage_height_offset)), name=f"Stage {current_stage.stage_number}")
            previous_stage = current_stage
        return assembly



from dataclasses import dataclass
from functools import cached_property
from typing import Optional, Union
import plotly.graph_objects as go
import numpy as np
from turbodesigner.airfoils.common import get_staggered_coords

@dataclass
class DCAAirfoil:
    c: float
    "chord length (length)"

    theta: float
    "camber angle (rad)"

    r0: float
    "double circular arc airfoil nose radius (length)"

    tb: float
    "max thickness (length)"

    xi: float = 0
    "stagger angle (rad)"

    arc_weight: float = 0.8
    "percentage of how much of x coordinates to load for arc"

    def __post_init__(self):
        if self.theta == 0:
            self.theta = 1E-5
        self.theta_mag = np.abs(self.theta)


    def get_camber_line(self, xc:Optional[Union[float, np.ndarray]] = None, is_staggered: bool = True, is_centered: bool = True,  num_points = 20):
        """coordinates of camber line (length)

            num_points: int
                number of points
        """
        # horizontal position of camber or chord line (length)
        if xc is None:
            xc = np.linspace(-self.c/2, self.c/2, num_points, endpoint=True)
        y_sign = np.sign(self.theta)
        # radius of curvature (length)
        Rc = (self.c/2) * (1/(np.sin(self.theta/2)))
        
        # camber line y-coordinate origin of radius (length)
        yc0 = -Rc * np.cos(self.theta/2)
        
        # vertical position of camber of chord line (length)
        yc = yc0 + np.sqrt(Rc**2 - xc**2) * y_sign

        camber_line = np.array([xc,yc]).T

        if is_centered:
            camber_line = camber_line - np.array([0,yc0 + np.sqrt(Rc**2) * y_sign])

        if is_staggered:
            camber_line = get_staggered_coords(camber_line, self.xi)
        return camber_line

    def get_circle(self, is_left: bool, num_points: int):
        """coordinates of circle for DCA airfoil (length)
        
            is_left: bool
                whether it is the left circle of DCA airfoil

            num_points: int
                number of points
        """

        x_sign = -1 if is_left else 1
        x_center = x_sign*(self.c/2 - self.r0*np.cos(self.theta_mag/2))
        y_center = self.r0 * np.sin(self.theta_mag/2)

        center_to_end_distance = (x_sign*self.c/2) - x_center
        angle_offset = np.pi - np.arccos(center_to_end_distance/self.r0) + np.pi/2
        angle = np.linspace(0, np.pi, num_points, endpoint=True) + angle_offset

        x = self.r0 * np.cos(angle) + x_center
        y = (self.r0 * np.sin(angle) + y_center) * np.sign(self.theta)

        return np.array([x,y]).T


    def get_arc(self, is_lower = False, num_points=20):
        """y-coordinates of chord line (length)
        
            is_lower: bool = False
                whether arc is lower
        
        """
        # negate tb and r0 if this a lower arc
        input_sign = -1 if is_lower else 1
        r0 = input_sign*self.r0
        tb = input_sign*self.tb

        x = np.linspace(-self.c*self.arc_weight/2, self.c*self.arc_weight/2, num=num_points)

        # camberline coordinate at mid coord (length)
        ym = (self.c/2)*np.tan(self.theta_mag/4)
        
        # radius of circular arc (length)
        d = ym + (tb / 2) - r0*np.sin(self.theta_mag/2)
        R = (d**2 - (r0**2) + ((self.c/2) - r0 * np.cos(self.theta_mag/2))**2)/(2*(d-r0))
        
        # origin of circular arc (length)
        y0 = ym + (tb/2) - R

        y = np.sign(self.theta) * (y0 + np.sqrt(R**2 - x**2))

        if is_lower and np.abs(y[0]) > self.tb*2:
            y = -np.sign(self.theta) * np.ones(num_points) * self.r0

        return np.array([x,y]).T
    
    def get_coords(
        self, 
        num_arc_points: int = 20, 
        num_circle_points: int = 10
    ):
        """double circular arc airfoil coordinates

            num_arc_points: int
                number of arc points

            num_circle_points: int
                number of circle points

        """
        # Info: x coordinates are [:,0] and y coordinates are [:,1]

        # Circles
        left_circle = self.get_circle(is_left=True, num_points=num_circle_points)
        right_circle = self.get_circle(is_left=False, num_points=num_circle_points)

        # Arcs
        upper_arc = self.get_arc(is_lower=False, num_points=num_arc_points)
        lower_arc = self.get_arc(is_lower=True, num_points=num_arc_points)
        left_circle_start = np.array([left_circle[0]])

        upper_cond = np.where(np.logical_and(upper_arc[:,0] > left_circle[0,0], upper_arc[:,0] < right_circle[-1,0]))
        lower_cond = np.where(np.logical_and(lower_arc[:,0] > left_circle[-1,0], lower_arc[:,0] < right_circle[0,0]))
        
        center = self.get_camber_line(0, is_staggered=False, is_centered=False)
        airfoil = np.concatenate(
            [
                left_circle, 
                lower_arc[lower_cond],
                right_circle, 
                np.flip(upper_arc[upper_cond], axis=0), 
                left_circle_start
            ]
        ) - center

        return get_staggered_coords(airfoil, self.xi)

    def visualize(
        self,
        fig=go.Figure(),
        show=True,
        num_arc_points: int = 20, 
        num_circle_points: int = 10
    ):
        camber_xy = self.get_camber_line(num_points=num_arc_points)
        fig.add_trace(go.Scatter(
            x=camber_xy[:, 0],
            y=camber_xy[:, 1],
        ))

        xy = self.get_coords(num_arc_points, num_circle_points)
        fig.add_trace(go.Scatter(
            x=xy[:, 0],
            y=xy[:, 1],
            fill="toself"
        ))

        fig.layout.yaxis.scaleanchor="x"  # type: ignore
        if show:
            fig.show()



from functools import cached_property
from dataclasses import dataclass
from typing import Optional
import numpy as np
from turbodesigner.blade.row import BladeRow, BladeRowCadExport, MetalAngleMethods
from turbodesigner.blade.vortex.free_vortex import FreeVortex
from turbodesigner.flow_station import FlowStation
from turbodesigner.units import MM


@dataclass
class StageBladeProperty:
    rotor: float
    stator: float


@dataclass
class StageCadExport:
    rotor: BladeRowCadExport
    "rotor blade row"

    stator: BladeRowCadExport
    "stator blade row"

    stage_height: float
    "stage height (mm)"

    stage_number: int
    "stage number"

    row_gap: float
    "stage gap (mm)"

    stage_gap: float
    "stage gap (mm)"


@dataclass
class Stage:
    "calculates turbomachinery stage"

    stage_number: int
    "stage number"

    Delta_Tt: float
    "stage stagnation temperature change between inlet and outlet (K)"

    R: float
    "stage reaction (dimensionless)"

    previous_flow_station: FlowStation
    "previous flow station (FlowStation)"

    eta_poly: float
    "polytropic efficiency (dimensionless)"

    N_stream: int
    "number of streams per blade (dimensionless)"

    AR: StageBladeProperty
    "aspect ratio (dimensionless)"

    sc: StageBladeProperty
    "spacing to chord ratio (dimensionless)"

    tbc: StageBladeProperty
    "max thickness to chord (dimensionless)"

    rgc: float
    "row gap to chord (dimensionless)"

    sgc: float
    "stage gap to chord (dimensionless)"

    metal_angle_method: MetalAngleMethods
    "metal angle method"

    next_stage: Optional["Stage"] = None
    "next turbomachinery stage"

    def __post_init__(self):
        assert isinstance(self.previous_flow_station.radius, float)
        self.rm = self.previous_flow_station.radius
        self.N = self.previous_flow_station.N

    @cached_property
    def Delta_ht(self) -> float:
        "enthalpy change between inlet and outlet (J/kg)"
        return self.Delta_Tt*self.previous_flow_station.Cp

    @cached_property
    def U(self):
        "mean blade velocity (m/s)"
        U = FlowStation.calc_U(self.N, self.rm)
        assert isinstance(U, float)
        return U

    @cached_property
    def phi(self):
        "flow coefficient (dimensionless)"
        return self.previous_flow_station.Vm/self.U

    @cached_property
    def psi(self):
        "loading coefficient (dimensionless)"
        return self.Delta_ht/self.U**2

    @cached_property
    def Tt2(self):
        "outlet total temperature (K)"
        return self.inlet_flow_station.Tt + self.Delta_Tt

    @cached_property
    def inlet_flow_station(self):
        "mid flow station between rotor and stator (FlowStation)"
        alpha1 = np.arctan((1 - self.R + -(1/2)*self.psi)/self.phi)
        return self.previous_flow_station.copyFlow(
            alpha=alpha1,
        )

    @cached_property
    def mid_flow_station(self):
        "mid flow station between rotor and stator (FlowStation)"
        alpha2 = np.arctan((1 - self.R + (1/2)*self.psi)/self.phi)
        Pt2 = self.inlet_flow_station.Pt*self.PR
        return self.inlet_flow_station.copyFlow(
            Tt=self.Tt2,
            Pt=Pt2,
            alpha=alpha2,
        )

    @cached_property
    def PR(self):
        "pressure ratio (dimensionless)"
        gamma = self.inlet_flow_station.gamma
        return self.TR**(self.eta_poly*gamma/(gamma - 1))

    @cached_property
    def TR(self):
        "stagnation temperature ratio between stage outlet and inlet (dimensionless)"
        return self.Tt2/self.inlet_flow_station.Tt

    @cached_property
    def tau(self):
        "torque transmitted to stage (N*m)"
        return self.inlet_flow_station.mdot*self.rm*(self.mid_flow_station.Vtheta - self.inlet_flow_station.Vtheta)

    @cached_property
    def vortex(self):
        return FreeVortex(
            Um=self.U,
            Vm=self.previous_flow_station.Vm,
            Rm=self.R,
            psi_m=self.psi,
            rm=self.rm
        )

    @cached_property
    def rotor(self):
        return BladeRow(
            stage_number=self.stage_number,
            stage_flow_station=self.inlet_flow_station,
            vortex=self.vortex,
            AR=self.AR.rotor,
            sc=self.sc.rotor,
            tbc=self.tbc.rotor,
            is_rotating=True,
            N_stream=self.N_stream,
            metal_angle_method=self.metal_angle_method,
            next_stage_flow_station=self.stator.flow_station
        )

    @cached_property
    def stator(self):
        return BladeRow(
            stage_number=self.stage_number,
            stage_flow_station=self.mid_flow_station,
            vortex=self.vortex,
            AR=self.AR.stator,
            sc=self.sc.stator,
            tbc=self.tbc.stator,
            is_rotating=False,
            N_stream=self.N_stream,
            metal_angle_method=self.metal_angle_method,
            next_stage_flow_station=None if self.next_stage is None else self.next_stage.rotor.flow_station
        )

    def to_cad_export(self):
        rotor = self.rotor.to_cad_export()
        stator = self.stator.to_cad_export()
        stage_height = rotor.disk_height+stator.disk_height
        return StageCadExport(
            stage_number=self.stage_number,
            rotor=rotor,
            stator=stator,
            stage_height=stage_height,
            stage_gap=self.sgc*self.rotor.c*MM,
            row_gap=self.rgc*self.rotor.c*MM
        )
